#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `pyuwsgimemhog` package."""

import pytest

from click.testing import CliRunner

from pyuwsgimemhog import pyuwsgimemhog
from pyuwsgimemhog import cli

@pytest.fixture
def invalid_sample_log():
    return ['invalid']


@pytest.fixture
def normalization_log():
    return """{address space usage: 806559744 bytes/769MB} {rss usage: 342519808 bytes/326MB} [pid: 4193|app: 0|req: 990967/3250990] 127.0.0.2 () {48 vars in 984 bytes} [Thu Apr 11 20:57:06 2019] POST /api/12 => generated 192 bytes in 71 msecs (HTTP/1.1 201) 5 headers in 153 bytes (1 switches on core 0)
{address space usage: 806559744 bytes/769MB} {rss usage: 352321536 bytes/336MB} [pid: 4193|app: 0|req: 990967/3250990] 127.0.0.2 () {48 vars in 984 bytes} [Thu Apr 11 20:57:06 2019] POST /api/12/ => generated 192 bytes in 71 msecs (HTTP/1.1 201) 5 headers in 153 bytes (1 switches on core 0)
{address space usage: 806559744 bytes/769MB} {rss usage: 354418688 bytes/336MB} [pid: 4193|app: 0|req: 990967/3250990] 127.0.0.2 () {48 vars in 984 bytes} [Thu Apr 11 20:57:06 2019] POST /api/11/ => generated 192 bytes in 71 msecs (HTTP/1.1 201) 5 headers in 153 bytes (1 switches on core 0)""".split('\n')


@pytest.fixture
def memory_reduction_log():
    return """{address space usage: 806559744 bytes/769MB} {rss usage: 342519808 bytes/326MB} [pid: 4193|app: 0|req: 990962/3250981] 127.0.0.1 () {36 vars in 573 bytes} [Thu Apr 11 20:57:06 2019] GET / => generated 23111 bytes in 2 msecs (HTTP/1.1 200) 5 headers in 165 bytes (1 switches on core 0)
{address space usage: 806559744 bytes/769MB} {rss usage: 283852800 bytes/270MB} [pid: 4193|app: 0|req: 990963/3250982] 127.0.0.1 () {36 vars in 573 bytes} [Thu Apr 11 20:57:06 2019] GET / => generated 23111 bytes in 1 msecs (HTTP/1.1 200) 5 headers in 165 bytes (1 switches on core 0)""".split('\n')


@pytest.fixture
def sample_log():
    return """{address space usage: 806559744 bytes/769MB} {rss usage: 342519808 bytes/326MB} [pid: 4193|app: 0|req: 990962/3250981] 127.0.0.1 () {36 vars in 573 bytes} [Thu Apr 11 20:57:06 2019] GET / => generated 23111 bytes in 2 msecs (HTTP/1.1 200) 5 headers in 165 bytes (1 switches on core 0)
{address space usage: 806559744 bytes/769MB} {rss usage: 346030080 bytes/330MB} [pid: 4193|app: 0|req: 990963/3250982] 127.0.0.1 () {36 vars in 573 bytes} [Thu Apr 11 20:57:06 2019] GET / => generated 23111 bytes in 1 msecs (HTTP/1.1 200) 5 headers in 165 bytes (1 switches on core 0)
{address space usage: 745689088 bytes/711MB} {rss usage: 282902528 bytes/269MB} [pid: 30820|app: 0|req: 592444/3250988] 127.0.0.1 () {36 vars in 573 bytes} [Thu Apr 11 20:57:06 2019] GET / => generated 23111 bytes in 2 msecs (HTTP/1.1 200) 5 headers in 165 bytes (1 switches on core 0)
{address space usage: 745689088 bytes/711MB} {rss usage: 283115520 bytes/270MB} [pid: 30820|app: 0|req: 592444/3250988] 127.0.0.1 () {36 vars in 573 bytes} [Thu Apr 11 20:57:06 2019] GET / => generated 23111 bytes in 2 msecs (HTTP/1.1 200) 5 headers in 165 bytes (1 switches on core 0)""".split('\n')


def test_uwsgimemhog_skips_invalid_lines(invalid_sample_log):
    data = pyuwsgimemhog.uwsgimemhog(invalid_sample_log, threshold=1, normalize=pyuwsgimemhog.normalize_path_with_nums)
    with pytest.raises(StopIteration):
        next(data)


def test_uwsgimemhog_normalize_paths_correctly(normalization_log):
    data = pyuwsgimemhog.uwsgimemhog(normalization_log, threshold=1,  normalize=pyuwsgimemhog.normalize_path)
    assert next(data) == ('/api/12', 9801728, 1)
    assert next(data) == ('/api/11', 2097152, 1)
    with pytest.raises(StopIteration):
        next(data)


def test_uwsgimemhog_normalize_nums_correctly(normalization_log):
    data = pyuwsgimemhog.uwsgimemhog(normalization_log, threshold=1, normalize=pyuwsgimemhog.normalize_path_with_nums)
    assert next(data) == ('/api/0', 11898880, 2)
    with pytest.raises(StopIteration):
        next(data)


def test_uwsgimemhog_ignores_negative_rss_diffs(memory_reduction_log):
    data = pyuwsgimemhog.uwsgimemhog(memory_reduction_log, threshold=1, normalize=pyuwsgimemhog.normalize_path_with_nums)
    with pytest.raises(StopIteration):
        next(data)


def test_uwsgimemhog_sum_same_path_on_different_pids_correctly(sample_log):
    data = pyuwsgimemhog.uwsgimemhog(sample_log, threshold=1, normalize=pyuwsgimemhog.normalize_path_with_nums)
    assert next(data) == ('/', 3723264, 2)
    with pytest.raises(StopIteration):
        next(data)


def test_uwsgimemhog_threshold_works_correctly(sample_log):
    data = pyuwsgimemhog.uwsgimemhog(sample_log, threshold=10_000_000, normalize=pyuwsgimemhog.normalize_path_with_nums)
    with pytest.raises(StopIteration):
        next(data)


def test_command_line_interface_renders_help():
    """Test the CLI."""
    runner = CliRunner()
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert 'Show this message and exit.' in help_result.output



# -*- coding: utf-8 -*-

"""Console script for pyuwsgimemhog."""
import sys
import click

from .pyuwsgimemhog import normalize_path, normalize_path_with_nums, uwsgimemhog


@click.command()
@click.option('--logfile', help='uWSGI log path', required=True)
@click.option('-m', '--memory-threshold', 'threshold',
              type=int, default=10, show_default=True,
              help='Memory threshold in MegaBytes')
@click.option('--normalize-nums/--no-normalize-nums', 'normalize_nums',
              default=True, show_default=True,
              help='Normalize numbers in urls')
def main(logfile, threshold, normalize_nums):
    """Console script for pyuwsgimemhog."""
    normalize = normalize_path_with_nums if normalize_nums else normalize_path
    with open(logfile, "r") as f:
        hogs = uwsgimemhog(f, threshold * 1_000_000, normalize)
        for path, memory, count in hogs:
            click.echo('{} {} {} {:.1f}'.format(
                path, memory // 1_000_000, count, memory / count / 1_000_000
            ))
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-
"""Main module."""
import re
from collections import defaultdict
from operator import itemgetter
from urllib.parse import urlparse


UWSGI_LOG_RE = re.compile(
    r'{rss usage: (?P<rss>\d+) bytes/\d+MB} \[pid: (?P<pid>\d+)\|app: .*\|req: .*\] '
    r'(?P<client_ip>\d+.\d+.\d+.\d+) .* \[(?P<date>\w+ \w+\s+\d+ \d+:\d+:\d+ \d+)\] '
    r'(?P<verb>GET|POST|HEAD|PATCH|PUT|DELETE) (?P<path>/.*) => '
    r'generated \d+ bytes in (?P<time>\d+) msecs \(HTTP/\d.\d (?P<status_code>\d+)\)'
)


def normalize_path(path):
    url = urlparse(path)
    return url.path.rstrip('/') if url.path != '/' else url.path


def normalize_path_with_nums(path):
    return re.sub(r'\d+', '0', normalize_path(path))


def uwsgimemhog(logfile, threshold, normalize=normalize_path):
    """Parse a uWSGI logfile and yields the RSS memory usage accumulated by each path"""
    pids_rss = defaultdict(list)
    for line in logfile:
        match = UWSGI_LOG_RE.search(line)
        if not match:
            continue

        path = normalize(match.group('path'))

        # store a list of path and rss for each pid
        pid = match.group('pid')
        rss = int(match.group('rss'))
        pids_rss[pid].append((path, rss))

    # accumulate the rss differences per path
    diffs = defaultdict(int)
    counter = defaultdict(int)
    for pid, v in pids_rss.items():
        for i in range(1, len(v)):
            _, prev_rss = v[i-1]
            path, rss = v[i]
            # don't track negatives value as we care about leaks
            rss_diff = rss - prev_rss
            if rss_diff < 0:
                continue
            diffs[path] += rss_diff
            counter[path] += 1

    # yield all paths that accumulated more than the threshold
    for path, value in sorted(diffs.items(), key=itemgetter(1), reverse=True):
        if value > threshold:
            yield path, value, counter[path]
